from typing import Optional, List, Any
import re


class ChineseTextSplitter:
    def __init__(
            self,
            chunk_size = 12,    #这里指的是每个
            chunk_overlap = 0,
            customer_clean_fun = None
    ) -> None:
        """Create a new TextSplitter."""
        super().__init__()

        print("\033[0;31;40m", "注意chunk_size定义为每个chunk中包含的字符个数，而不是chunk的数目", "\033[0m")
        # 确保chunk_size大于chunk_overlap
        assert chunk_size > chunk_overlap, "Chunk size must be greater than overlap."

        self._separators = [
            "\n\n",
            "\n",
            "[。！？]",  # 使用字符类来匹配中文字符的句号、感叹号和问号
            "\.\s|\!\s|\?\s",
            "[；;]\s",  # 同时匹配中文分号和英文分号
            "[，,]\s"  # 同时匹配中文逗号和英文逗号
        ]

        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.customer_clean_fun = customer_clean_fun

    def customer_clean(self,text):
        return self.customer_clean_fun(text)

    def load_txt(self,txt_file_path,need_clean_text = True):

        file_content = "".join(open(txt_file_path, "r", encoding="UTF-8").readlines())
        if need_clean_text:
            file_content = self.clean_text(file_content)

        if self.customer_clean_fun != None:
            file_content = self.customer_clean(file_content)


        chunk_size = self.chunk_size
        chunk_overlap = self.chunk_overlap

        # 初始化结果列表
        text_num = len(file_content)
        chunk_step = chunk_size - chunk_overlap
        chunks_num = text_num//chunk_step + 1
        print(f"您预期划分的每个段落块中有{chunk_size}个字符，预期划分的块数chunks_num为{chunks_num}个。总字数为{text_num}")

        return file_content


    def recursive_split_text(self, text: str,need_clean_text = True) -> List[str]:
        if need_clean_text:
            text = self.clean_text(text)

        if self.customer_clean_fun != None:
            file_content = self.customer_clean(text)


        chunk_size = self.chunk_size
        chunk_overlap = self.chunk_overlap

        # 初始化结果列表
        chunks = []
        text_num = len(text)
        chunk_step = chunk_size - chunk_overlap
        chunks_num = text_num//chunk_step + 1
        print(f"您预期划分的每个段落块中有{chunk_size}个字符，预期划分的块数chunks_num为{chunks_num}个。总字数为{text_num}")
        for i in range(chunks_num):
            start = i * chunk_step
            end = (i+1) * chunk_size
            chunks.append(text[start:end])

            # 返回结果列表
        return chunks

    def clean_text(self, text):
        separators = self._separators
        # 将包含管道符的模式字符串修正为正则表达式
        regex_patterns = [
                             re.escape(sep.replace('|', '')) for sep in separators if '|' not in sep
                         ] + [
                             sep.replace('|', '') for sep in separators if '|' in sep
                         ]
        # 合并所有的模式为一个大的正则表达式，使用管道符连接
        regex_combined = '|'.join(regex_patterns)
        # 使用正则表达式替换所有的匹配项为空字符串
        cleaned_text = re.sub(regex_combined, ' ', text)
        # 额外的步骤：将多个连续空格替换为单个空格，并去除文本开头和结尾的空格
        cleaned_text = re.sub(r'\s+', ' ', cleaned_text).strip()
        return cleaned_text



if __name__ == "__main__":
    text_splitter = ChineseTextSplitter(
        chunk_size=1280,
        chunk_overlap=5
    )

    #file_path = "./2020-02-21__中国平安保险集团股份有限公司__601318__中国平安__2019年__年度报告.txt"
    #file_content = "".join(open(file_path, "r", encoding="UTF-8").readlines())
    #file_content = text_splitter.clean_text(file_content)
    file_content = "本来我抽了几天晚上和周末时间，刚开始提交是17名左右，在上周末被更多分数冲下来了，卷得飞起，主要是还在开发项目，时间确实有限，本来以为自己很拼了，我粗略估计，别人花的时间是我的十倍以上，让我想起了那句话：别拿自己的爱好挑战别人吃饭的技能，虽然我的AI能力很强， 但是金融知识确实欠缺，我有个梦想，就是AI可以逐步理解行业的思考逻辑，或者叫行业knowhow，以后做行业相关的还得打牢基础，目前还有点远。之，感谢举办方、感谢小打小闹，感谢小姐姐，加我微信指导我，挺负责的。以上仅是我的理解，有不妥之处还望指出，我来修正。"

    chunks = text_splitter.recursive_split_text(file_content)





import warnings
warnings.filterwarnings("ignore")

import asyncio
from abc import ABC, abstractmethod
from typing import List


class Embeddings(ABC):
    """Interface for embedding models."""

    @abstractmethod
    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs."""

    @abstractmethod
    def embed_query(self, text: str) -> List[float]:
        """Embed query text."""

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Asynchronous Embed search docs."""
        return await asyncio.get_running_loop().run_in_executor(
            None, self.embed_documents, texts
        )

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous Embed query text."""
        return await asyncio.get_running_loop().run_in_executor(
            None, self.embed_query, text
        )


# class LMEmbedding(Embeddings):
#     from modelscope.pipelines import pipeline
#     from modelscope.utils.constant import Tasks
#     pipeline_se = pipeline(Tasks.sentence_embedding, model='thomas/text2vec-base-chinese', model_revision='v1.0.0',
#                            device="cuda")
#
#     def _costruct_inputs(self, texts):
#         inputs = {
#             "source_sentence": texts
#         }
#
#         return inputs
#
#     def embed_documents(self, texts: List[str]) -> List[List[float]]:
#         """Embed search docs."""
#
#         inputs = self._costruct_inputs(texts)
#         result_embeddings = self.pipeline_se(input=inputs)
#         return result_embeddings["text_embedding"]
#
#     def embed_query(self, text: str) -> List[float]:
#         """Embed query text."""
#         inputs = self._costruct_inputs([text])
#         result_embeddings = self.pipeline_se(input=inputs)
#         return result_embeddings["text_embedding"]


class GLMEmbedding(Embeddings):
    import zhipuai as zhipuai
    zhipuai.api_key = "1f565e40af1198e11ff1fd8a5b42771d.SjNfezc40YFsz2KC"  # 控制台中获取的 APIKey 信息

    def _costruct_inputs(self, texts):
        inputs = {
            "source_sentence": texts
        }

        return inputs

    aembeddings = []  # 这个是为了在并发获取embedding_value时候使用的存储embedding_list内容。
    atexts = []

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs."""
        result_embeddings = []
        for text in texts:
            embedding = self.embed_query(text)
            result_embeddings.append(embedding)
        return result_embeddings

    def embed_query(self, text: str) -> List[float]:
        """Embed query text."""
        result_embeddings = self.zhipuai.model_api.invoke(
            model="text_embedding", prompt=text)
        return result_embeddings["data"]["embedding"]

    def aembed_query(self, text: str) -> List[float]:
        """Embed query text."""
        result_embeddings = self.zhipuai.model_api.invoke(
            model="text_embedding", prompt=text)
        emb = result_embeddings["data"]["embedding"]

        self.aembeddings.append(emb)
        self.atexts.append(text)

    # 这里实现了并发embedding获取
    def aembed_documents(self, texts: List[str], thread_num=3, wait_sec=0.4) -> List[List[float]]:
        import threading
        text_length = len(texts)
        thread_batch = text_length // thread_num

        for i in range(thread_batch):
            start = i * thread_num
            end = (i + 1) * thread_num

            # 创建线程列表
            threads = []
            # 创建并启动5个线程，每个线程调用一个模型
            for text in texts[start:end]:
                thread = threading.Thread(target=self.aembed_query, args=(text,))
                thread.start()
                threads.append(thread)
            for thread in threads:
                thread.join(wait_sec)  # 设置超时时间为0.3秒
        return self.aembeddings, self.atexts


if __name__ == '__main__':
    import time

    inputs = ["不可以，早晨喝牛奶不科学", "今天早晨喝牛奶不科学", "早晨喝牛奶不科学"] * 50

    start_time = time.time()
    aembeddings = (GLMEmbedding().aembed_documents(inputs, thread_num=5, thread_sec=0.3))
    print(aembeddings)
    print(len(aembeddings))
    end_time = time.time()
    # 计算函数执行时间并打印结果
    execution_time = end_time - start_time
    print(f"函数执行时间: {execution_time} 秒")
    print("----------------------------------------------------------------------------------")
    start_time = time.time()
    aembeddings = (GLMEmbedding().embed_documents(inputs))
    print(len(aembeddings))
    end_time = time.time()
    # 计算函数执行时间并打印结果
    execution_time = end_time - start_time
    print(f"函数执行时间: {execution_time} 秒")



from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate

from tqdm import tqdm
from lmchain.tools import tool_register


class SubQuestChain:
    def __init__(self, llm):
        self.llm = llm

    def __call__(self, query=""):
        if query == "":
            raise "query需要填入查询问题"

        decomp_template = """
            GENERAL INSTRUCTIONS
            You are a domain expert. Your task is to break down a complex question into simpler sub-parts.

            USER QUESTION
            {user_question}

            ANSWER FORMAT
            ["sub-questions_1","sub-questions_2","sub-questions_3",...]
            """

        from langchain.prompts import PromptTemplate
        prompt = PromptTemplate(
            input_variables=["user_question"],
            template=decomp_template,
        )

        from langchain.chains import LLMChain
        chain = LLMChain(llm=self.llm, prompt=prompt)
        response = (chain.run({"user_question": query}))

        import json
        sub_list = json.loads(response)

        return sub_list

    def run(self, query):
        sub_list = self.__call__(query)
        return sub_list


if __name__ == '__main__':
    from lmchain.agents import llmMultiAgent

    llm = llmMultiAgent.AgentZhipuAI()

    subQC = SubQuestChain(llm)
    response = subQC.run(query="工商银行财报中，2024财年Q1与Q2 之间，利润增长了多少？")
    print(response)


from __future__ import annotations

import enum
import threading
from abc import abstractmethod
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)
from weakref import WeakValueDictionary

from langchain.pydantic_v1 import BaseModel
from langchain.schema.runnable.base import Runnable, RunnableSerializable
from langchain.schema.runnable.config import (
    RunnableConfig,
    get_config_list,
    get_executor_for_config,
)
from langchain.schema.runnable.utils import (
    AnyConfigurableField,
    ConfigurableField,
    ConfigurableFieldMultiOption,
    ConfigurableFieldSingleOption,
    ConfigurableFieldSpec,
    Input,
    Output,
    gather_with_concurrency,
)


class DynamicRunnable(RunnableSerializable[Input, Output]):
    """A Serializable Runnable that can be dynamically configured."""

    default: RunnableSerializable[Input, Output]

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def is_lc_serializable(cls) -> bool:
        return True

    @classmethod
    def get_lc_namespace(cls) -> List[str]:
        return cls.__module__.split(".")[:-1]

    @property
    def InputType(self) -> Type[Input]:
        return self.default.InputType

    @property
    def OutputType(self) -> Type[Output]:
        return self.default.OutputType

    def get_input_schema(
        self, config: Optional[RunnableConfig] = None
    ) -> Type[BaseModel]:
        return self._prepare(config).get_input_schema(config)

    def get_output_schema(
        self, config: Optional[RunnableConfig] = None
    ) -> Type[BaseModel]:
        return self._prepare(config).get_output_schema(config)

    @abstractmethod
    def _prepare(
        self, config: Optional[RunnableConfig] = None
    ) -> Runnable[Input, Output]:
        ...

    def invoke(
        self, input: Input, config: Optional[RunnableConfig] = None, **kwargs: Any
    ) -> Output:
        return self._prepare(config).invoke(input, config, **kwargs)

    async def ainvoke(
        self, input: Input, config: Optional[RunnableConfig] = None, **kwargs: Any
    ) -> Output:
        return await self._prepare(config).ainvoke(input, config, **kwargs)

    def batch(
        self,
        inputs: List[Input],
        config: Optional[Union[RunnableConfig, List[RunnableConfig]]] = None,
        *,
        return_exceptions: bool = False,
        **kwargs: Optional[Any],
    ) -> List[Output]:
        configs = get_config_list(config, len(inputs))
        prepared = [self._prepare(c) for c in configs]

        if all(p is self.default for p in prepared):
            return self.default.batch(
                inputs, config, return_exceptions=return_exceptions, **kwargs
            )

        if not inputs:
            return []

        configs = get_config_list(config, len(inputs))

        def invoke(
            bound: Runnable[Input, Output],
            input: Input,
            config: RunnableConfig,
        ) -> Union[Output, Exception]:
            if return_exceptions:
                try:
                    return bound.invoke(input, config, **kwargs)
                except Exception as e:
                    return e
            else:
                return bound.invoke(input, config, **kwargs)

        # If there's only one input, don't bother with the executor
        if len(inputs) == 1:
            return cast(List[Output], [invoke(prepared[0], inputs[0], configs[0])])

        with get_executor_for_config(configs[0]) as executor:
            return cast(
                List[Output], list(executor.map(invoke, prepared, inputs, configs))
            )

    async def abatch(
        self,
        inputs: List[Input],
        config: Optional[Union[RunnableConfig, List[RunnableConfig]]] = None,
        *,
        return_exceptions: bool = False,
        **kwargs: Optional[Any],
    ) -> List[Output]:
        configs = get_config_list(config, len(inputs))
        prepared = [self._prepare(c) for c in configs]

        if all(p is self.default for p in prepared):
            return await self.default.abatch(
                inputs, config, return_exceptions=return_exceptions, **kwargs
            )

        if not inputs:
            return []

        configs = get_config_list(config, len(inputs))

        async def ainvoke(
            bound: Runnable[Input, Output],
            input: Input,
            config: RunnableConfig,
        ) -> Union[Output, Exception]:
            if return_exceptions:
                try:
                    return await bound.ainvoke(input, config, **kwargs)
                except Exception as e:
                    return e
            else:
                return await bound.ainvoke(input, config, **kwargs)

        coros = map(ainvoke, prepared, inputs, configs)
        return await gather_with_concurrency(configs[0].get("max_concurrency"), *coros)

    def stream(
        self,
        input: Input,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> Iterator[Output]:
        return self._prepare(config).stream(input, config, **kwargs)

    async def astream(
        self,
        input: Input,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> AsyncIterator[Output]:
        async for chunk in self._prepare(config).astream(input, config, **kwargs):
            yield chunk

    def transform(
        self,
        input: Iterator[Input],
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> Iterator[Output]:
        return self._prepare(config).transform(input, config, **kwargs)

    async def atransform(
        self,
        input: AsyncIterator[Input],
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> AsyncIterator[Output]:
        async for chunk in self._prepare(config).atransform(input, config, **kwargs):
            yield chunk


class RunnableConfigurableFields(DynamicRunnable[Input, Output]):
    """A Runnable that can be dynamically configured."""

    fields: Dict[str, AnyConfigurableField]

    @property
    def config_specs(self) -> Sequence[ConfigurableFieldSpec]:
        return [
            ConfigurableFieldSpec(
                id=spec.id,
                name=spec.name,
                description=spec.description
                or self.default.__fields__[field_name].field_info.description,
                annotation=spec.annotation
                or self.default.__fields__[field_name].annotation,
                default=getattr(self.default, field_name),
            )
            if isinstance(spec, ConfigurableField)
            else make_options_spec(
                spec, self.default.__fields__[field_name].field_info.description
            )
            for field_name, spec in self.fields.items()
        ]

    def configurable_fields(
        self, **kwargs: AnyConfigurableField
    ) -> RunnableSerializable[Input, Output]:
        return self.default.configurable_fields(**{**self.fields, **kwargs})

    def _prepare(
        self, config: Optional[RunnableConfig] = None
    ) -> Runnable[Input, Output]:
        config = config or {}
        specs_by_id = {spec.id: (key, spec) for key, spec in self.fields.items()}
        configurable_fields = {
            specs_by_id[k][0]: v
            for k, v in config.get("configurable", {}).items()
            if k in specs_by_id and isinstance(specs_by_id[k][1], ConfigurableField)
        }
        configurable_single_options = {
            k: v.options[(config.get("configurable", {}).get(v.id) or v.default)]
            for k, v in self.fields.items()
            if isinstance(v, ConfigurableFieldSingleOption)
        }
        configurable_multi_options = {
            k: [
                v.options[o]
                for o in config.get("configurable", {}).get(v.id, v.default)
            ]
            for k, v in self.fields.items()
            if isinstance(v, ConfigurableFieldMultiOption)
        }
        configurable = {
            **configurable_fields,
            **configurable_single_options,
            **configurable_multi_options,
        }

        if configurable:
            return self.default.__class__(**{**self.default.__dict__, **configurable})
        else:
            return self.default


# Before Python 3.11 native StrEnum is not available
class StrEnum(str, enum.Enum):
    """A string enum."""

    pass


_enums_for_spec: WeakValueDictionary[
    Union[
        ConfigurableFieldSingleOption, ConfigurableFieldMultiOption, ConfigurableField
    ],
    Type[StrEnum],
] = WeakValueDictionary()

_enums_for_spec_lock = threading.Lock()


class RunnableConfigurableAlternatives(DynamicRunnable[Input, Output]):
    """A Runnable that can be dynamically configured."""

    which: ConfigurableField

    alternatives: Dict[
        str,
        Union[Runnable[Input, Output], Callable[[], Runnable[Input, Output]]],
    ]

    default_key: str = "default"

    @property
    def config_specs(self) -> Sequence[ConfigurableFieldSpec]:
        with _enums_for_spec_lock:
            if which_enum := _enums_for_spec.get(self.which):
                pass
            else:
                which_enum = StrEnum(  # type: ignore[call-overload]
                    self.which.name or self.which.id,
                    (
                        (v, v)
                        for v in list(self.alternatives.keys()) + [self.default_key]
                    ),
                )
                _enums_for_spec[self.which] = cast(Type[StrEnum], which_enum)
        return [
            ConfigurableFieldSpec(
                id=self.which.id,
                name=self.which.name,
                description=self.which.description,
                annotation=which_enum,
                default=self.default_key,
            ),
            *self.default.config_specs,
        ] + [
            s
            for alt in self.alternatives.values()
            if isinstance(alt, RunnableSerializable)
            for s in alt.config_specs
        ]

    def configurable_fields(
        self, **kwargs: AnyConfigurableField
    ) -> RunnableSerializable[Input, Output]:
        return self.__class__(
            which=self.which,
            default=self.default.configurable_fields(**kwargs),
            alternatives=self.alternatives,
        )

    def _prepare(
        self, config: Optional[RunnableConfig] = None
    ) -> Runnable[Input, Output]:
        config = config or {}
        which = config.get("configurable", {}).get(self.which.id, self.default_key)
        if which == self.default_key:
            return self.default
        elif which in self.alternatives:
            alt = self.alternatives[which]
            if isinstance(alt, Runnable):
                return alt
            else:
                return alt()
        else:
            raise ValueError(f"Unknown alternative: {which}")


def make_options_spec(
    spec: Union[ConfigurableFieldSingleOption, ConfigurableFieldMultiOption],
    description: Optional[str],
) -> ConfigurableFieldSpec:
    """Make a ConfigurableFieldSpec for a ConfigurableFieldSingleOption or
    ConfigurableFieldMultiOption."""
    with _enums_for_spec_lock:
        if enum := _enums_for_spec.get(spec):
            pass
        else:
            enum = StrEnum(  # type: ignore[call-overload]
                spec.name or spec.id,
                ((v, v) for v in list(spec.options.keys())),
            )
            _enums_for_spec[spec] = cast(Type[StrEnum], enum)
    if isinstance(spec, ConfigurableFieldSingleOption):
        return ConfigurableFieldSpec(
            id=spec.id,
            name=spec.name,
            description=spec.description or description,
            annotation=enum,
            default=spec.default,
        )
    else:
        return ConfigurableFieldSpec(
            id=spec.id,
            name=spec.name,
            description=spec.description or description,
            annotation=Sequence[enum],  # type: ignore[valid-type]
            default=spec.default,
        )
